@@ -6,13 +6,13 @@ import os
 import re
 import sys
 from collections import defaultdict
-from dataclasses import field
+from enum import Enum
 from os import environ as env
 from os.path import dirname
-from typing import Any, Callable, Dict, List, Optional, Type, Union, cast
+from typing import Any, Callable, Dict, List, Optional, Sequence, Type, Union, cast
 from urllib.parse import urlparse
 
-from pydantic import Field, validator
+from pydantic import validator
 from pydantic.dataclasses import dataclass
 from pydantic.json import pydantic_encoder
 from ruamel.yaml import YAML
@@ -29,6 +29,11 @@ sys.path.append(os.getcwd())
 _logger = logging.getLogger(__name__)
 
 
+class OperationType(Enum):
+    transaction = 'transaction'
+    origination = 'origination'
+
+
 @dataclass
 class SqliteDatabaseConfig:
     """
@@ -142,27 +147,42 @@ class TzktDatasourceConfig:
 
 
 @dataclass
-class OperationHandlerPatternConfig:
+class OperationHandlerTransactionPatternConfig:
     """Operation handler pattern config
 
     :param destination: Alias of the contract to match
     :param entrypoint: Contract entrypoint
     """
 
-    destination: Union[str, ContractConfig]
-    entrypoint: str
+    type: Literal['transaction'] = 'transaction'
+    source: Optional[Union[str, ContractConfig]] = None
+    destination: Optional[Union[str, ContractConfig]] = None
+    entrypoint: Optional[str] = None
+    optional: bool = False
 
     def __post_init_post_parse__(self):
+        if self.entrypoint and not self.destination:
+            raise ConfigurationError('Transactions with entrypoint must also have destination')
         self._parameter_type_cls = None
         self._storage_type_cls = None
+        self._transaction_id = None
 
     @property
-    def contract_config(self) -> ContractConfig:
-        assert isinstance(self.destination, ContractConfig)
+    def source_contract_config(self) -> ContractConfig:
+        if not isinstance(self.source, ContractConfig):
+            raise RuntimeError('Config is not initialized')
+        return self.source
+
+    @property
+    def destination_contract_config(self) -> ContractConfig:
+        if not isinstance(self.destination, ContractConfig):
+            raise RuntimeError('Config is not initialized')
         return self.destination
 
     @property
-    def parameter_type_cls(self) -> Type:
+    def parameter_type_cls(self) -> Optional[Type]:
+        if not self.entrypoint:
+            raise RuntimeError('entrypoint is empty')
         if self._parameter_type_cls is None:
             raise RuntimeError('Config is not initialized')
         return self._parameter_type_cls
@@ -173,6 +193,8 @@ class OperationHandlerPatternConfig:
 
     @property
     def storage_type_cls(self) -> Type:
+        if not self.entrypoint:
+            raise RuntimeError('entrypoint is empty')
         if self._storage_type_cls is None:
             raise RuntimeError('Config is not initialized')
         return self._storage_type_cls
@@ -181,6 +203,87 @@ class OperationHandlerPatternConfig:
     def storage_type_cls(self, typ: Type) -> None:
         self._storage_type_cls = typ
 
+    @property
+    def transaction_id(self) -> int:
+        if self._transaction_id is None:
+            raise RuntimeError('Config is not initialized')
+        return self._transaction_id
+
+    @transaction_id.setter
+    def transaction_id(self, id_: int) -> None:
+        self._transaction_id = id_
+
+    def get_handler_imports(self, package: str) -> str:
+        if self.entrypoint:
+            module_name = self.destination_contract_config.module_name
+            entrypoint = camel_to_snake(self.entrypoint)
+            parameter_cls = f'{snake_to_camel(self.entrypoint)}Parameter'
+            storage_cls = f'{snake_to_camel(module_name)}Storage'
+            return '\n'.join(
+                [
+                    f'from {package}.types.{module_name}.parameter.{entrypoint} import {parameter_cls}',
+                    f'from {package}.types.{module_name}.storage import {storage_cls}',
+                ]
+            )
+        else:
+            return ''
+
+    def get_handler_argument(self) -> str:
+        if self.entrypoint:
+            module_name = self.destination_contract_config.module_name
+            entrypoint = camel_to_snake(self.entrypoint)
+            parameter_cls = f'{snake_to_camel(self.entrypoint)}Parameter'
+            storage_cls = f'{snake_to_camel(module_name)}Storage'
+            if self.optional:
+                return f'{entrypoint}: Optional[TransactionContext[{parameter_cls}, {storage_cls}]],'
+            return f'{entrypoint}: TransactionContext[{parameter_cls}, {storage_cls}],'
+        else:
+            if self.optional:
+                return f'transaction_{self._transaction_id}: Optional[OperationData],'
+            return f'transaction_{self._transaction_id}: OperationData,'
+
+
+@dataclass
+class OperationHandlerOriginationPatternConfig:
+    originated_contract: Union[str, ContractConfig]
+    type: Literal['origination'] = 'origination'
+    optional: bool = False
+
+    def __post_init_post_parse__(self):
+        self._storage_type_cls = None
+
+    @property
+    def parameter_type_cls(self) -> Optional[Type]:
+        return None
+
+    @property
+    def contract_config(self) -> ContractConfig:
+        if not isinstance(self.originated_contract, ContractConfig):
+            raise RuntimeError('Config is not initialized')
+        return self.originated_contract
+
+    @property
+    def storage_type_cls(self) -> Type:
+        if self._storage_type_cls is None:
+            raise RuntimeError('Config is not initialized')
+        return self._storage_type_cls
+
+    @storage_type_cls.setter
+    def storage_type_cls(self, typ: Type) -> None:
+        self._storage_type_cls = typ
+
+    def get_handler_imports(self, package: str) -> str:
+        module_name = self.contract_config.module_name
+        storage_cls = f'{snake_to_camel(module_name)}Storage'
+        return f'from {package}.types.{module_name}.storage import {storage_cls}'
+
+    def get_handler_argument(self) -> str:
+        module_name = self.contract_config.module_name
+        storage_cls = f'{snake_to_camel(module_name)}Storage'
+        if self.optional:
+            return f'{module_name}_origination: Optional[OriginationContext[{storage_cls}]],'
+        return f'{module_name}_origination: OriginationContext[{storage_cls}],'
+
 
 @dataclass
 class HandlerConfig:
@@ -200,6 +303,9 @@ class HandlerConfig:
         self._callback_fn = fn
 
 
+OperationHandlerPatternConfigT = Union[OperationHandlerOriginationPatternConfig, OperationHandlerTransactionPatternConfig]
+
+
 @dataclass
 class OperationHandlerConfig(HandlerConfig):
     """Operation handler config
@@ -208,7 +314,7 @@ class OperationHandlerConfig(HandlerConfig):
     :param pattern: Filters to match operations in group
     """
 
-    pattern: List[OperationHandlerPatternConfig]
+    pattern: List[OperationHandlerPatternConfigT]
 
 
 @dataclass
@@ -266,6 +372,7 @@ class OperationIndexConfig(IndexConfig):
     kind: Literal["operation"]
     contracts: List[Union[str, ContractConfig]]
     handlers: List[OperationHandlerConfig]
+    types: Optional[List[OperationType]] = None
     first_block: int = 0
     last_block: int = 0
 
@@ -342,14 +449,31 @@ class BlockIndexConfig(IndexConfig):
 
 
 @dataclass
-class IndexTemplateConfig:
+class StaticTemplateConfig:
     kind = 'template'
     template: str
     values: Dict[str, str]
 
 
-IndexConfigT = Union[OperationIndexConfig, BigMapIndexConfig, BlockIndexConfig, IndexTemplateConfig]
+@dataclass
+class DynamicTemplateConfig:
+    kind = 'dynamic'
+    template: str
+    similar_to: Union[str, ContractConfig]
+    strict: bool = False
+
+    @property
+    def contract_config(self) -> ContractConfig:
+        if not isinstance(self.similar_to, ContractConfig):
+            raise RuntimeError('Config is not initialized')
+        return self.similar_to
+
+
+IndexConfigT = Union[OperationIndexConfig, BigMapIndexConfig, BlockIndexConfig, StaticTemplateConfig, DynamicTemplateConfig]
 IndexConfigTemplateT = Union[OperationIndexConfig, BigMapIndexConfig, BlockIndexConfig]
+HandlerPatternConfigT = Union[
+    OperationHandlerOriginationPatternConfig, OperationHandlerTransactionPatternConfig, BigMapHandlerPatternConfig
+]
 
 
 @dataclass
@@ -389,12 +513,20 @@ class DipDupConfig:
     hasura: Optional[HasuraConfig] = None
 
     def __post_init_post_parse__(self):
+        self.validate()
+        self.pre_initialize()
+
+    def validate(self) -> None:
         if isinstance(self.database, SqliteDatabaseConfig) and self.hasura:
             raise ConfigurationError('SQLite DB engine is not supported by Hasura')
 
+    def pre_initialize(self) -> None:
         _logger.info('Substituting index templates')
         for index_name, index_config in self.indexes.items():
-            if isinstance(index_config, IndexTemplateConfig):
+            # NOTE: Dynamic templates will be resolved later in dipdup module
+            if isinstance(index_config, StaticTemplateConfig):
+                if not self.templates:
+                    raise ConfigurationError('`templates` section is missing')
                 try:
                     template = self.templates[index_config.template]
                 except KeyError as e:
@@ -404,10 +536,11 @@ class DipDupConfig:
                     value_regex = r'<[ ]*' + key + r'[ ]*>'
                     raw_template = re.sub(value_regex, value, raw_template)
                 json_template = json.loads(raw_template)
-                self.indexes[index_name] = template.__class__(**json_template)
-                self.indexes[index_name].template_values = index_config.values
+                new_index_config = template.__class__(**json_template)
+                new_index_config.template_values = index_config.values
+                self.indexes[index_name] = new_index_config
 
-        callback_patterns: Dict[str, List[List[OperationHandlerPatternConfig]]] = defaultdict(list)
+        callback_patterns: Dict[str, List[Sequence[HandlerPatternConfigT]]] = defaultdict(list)
 
         _logger.info('Substituting contracts and datasources')
         for index_config in self.indexes.values():
@@ -425,18 +558,44 @@ class DipDupConfig:
                         except KeyError as e:
                             raise ConfigurationError(f'Contract `{contract}` not found in `contracts` config section') from e
 
-                for handler in index_config.handlers:
-                    callback_patterns[handler.callback].append(handler.pattern)
-                    for pattern in handler.pattern:
-                        if isinstance(pattern.destination, str):
-                            try:
-                                pattern.destination = self.contracts[pattern.destination]
-                            except KeyError as e:
-                                raise ConfigurationError(f'Contract `{pattern.destination}` not found in `contracts` config section') from e
+                transaction_id = 0
+                for handler_config in index_config.handlers:
+                    callback_patterns[handler_config.callback].append(handler_config.pattern)
+                    for pattern_config in handler_config.pattern:
+                        if isinstance(pattern_config, OperationHandlerTransactionPatternConfig):
+                            if isinstance(pattern_config.destination, str):
+                                try:
+                                    pattern_config.destination = self.contracts[pattern_config.destination]
+                                except KeyError as e:
+                                    raise ConfigurationError(
+                                        f'Contract `{pattern_config.destination}` not found in `contracts` config section'
+                                    ) from e
+                            if isinstance(pattern_config.source, str):
+                                try:
+                                    pattern_config.source = self.contracts[pattern_config.source]
+                                except KeyError as e:
+                                    raise ConfigurationError(
+                                        f'Contract `{pattern_config.source}` not found in `contracts` config section'
+                                    ) from e
+                            if not pattern_config.entrypoint:
+                                pattern_config.transaction_id = transaction_id
+                                transaction_id += 1
+
+                        elif isinstance(pattern_config, OperationHandlerOriginationPatternConfig):
+                            if isinstance(pattern_config.originated_contract, str):
+                                try:
+                                    pattern_config.originated_contract = self.contracts[pattern_config.originated_contract]
+                                except KeyError as e:
+                                    raise ConfigurationError(
+                                        f'Contract `{pattern_config.originated_contract}` not found in `contracts` config section'
+                                    ) from e
 
             elif isinstance(index_config, BigMapIndexConfig):
                 if isinstance(index_config.datasource, str):
-                    index_config.datasource = self.datasources[index_config.datasource]
+                    try:
+                        index_config.datasource = self.datasources[index_config.datasource]
+                    except KeyError as e:
+                        raise ConfigurationError(f'Datasource `{index_config.datasource}` not found in `datasources` config section') from e
 
                 for handler in index_config.handlers:
                     callback_patterns[handler.callback].append(handler.pattern)
@@ -447,6 +606,10 @@ class DipDupConfig:
                             except KeyError as e:
                                 raise ConfigurationError(f'Contract `{pattern.contract}` not found in `contracts` config section') from e
 
+            # NOTE: Dynamic templates will be resolved later in dipdup module
+            elif isinstance(index_config, DynamicTemplateConfig):
+                continue
+
             else:
                 raise NotImplementedError(f'Index kind `{index_config.kind}` is not supported')
 
@@ -454,8 +617,15 @@ class DipDupConfig:
         for callback, patterns in callback_patterns.items():
             if len(patterns) > 1:
 
-                def get_pattern_type(pattern: List[OperationHandlerPatternConfig]):
-                    return '::'.join(map(lambda x: x.contract_config.module_name, pattern))
+                def get_pattern_type(pattern: Sequence[HandlerPatternConfigT]) -> str:
+                    module_names = []
+                    for pattern_config in pattern:
+                        if isinstance(pattern_config, OperationHandlerTransactionPatternConfig) and pattern_config.entrypoint:
+                            module_names.append(pattern_config.destination_contract_config.module_name)
+                        elif isinstance(pattern_config, OperationHandlerOriginationPatternConfig):
+                            module_names.append(pattern_config.contract_config.module_name)
+                        # TODO: Check BigMapHandlerPatternConfig
+                    return '::'.join(module_names)
 
                 pattern_types = list(map(get_pattern_type, patterns))
                 if any(map(lambda x: x != pattern_types[0], pattern_types)):
@@ -537,8 +707,11 @@ class DipDupConfig:
 
         for index_name, index_config in self.indexes.items():
 
-            if isinstance(index_config, IndexTemplateConfig):
+            if isinstance(index_config, StaticTemplateConfig):
                 raise RuntimeError('Config is not initialized')
+            # NOTE: Dynamic templates will be resolved later in dipdup module
+            if isinstance(index_config, DynamicTemplateConfig):
+                continue
 
             await self._initialize_index_state(index_name, index_config)
 
@@ -548,27 +721,42 @@ class DipDupConfig:
                     await self._initialize_handler_callback(operation_handler_config)
 
                     for operation_pattern_config in operation_handler_config.pattern:
-                        _logger.info('Registering parameter type for entrypoint `%s`', operation_pattern_config.entrypoint)
-                        parameter_type_module = importlib.import_module(
-                            f'{self.package}'
-                            f'.types'
-                            f'.{operation_pattern_config.contract_config.module_name}'
-                            f'.parameter'
-                            f'.{camel_to_snake(operation_pattern_config.entrypoint)}'
-                        )
-                        parameter_type_cls = getattr(
-                            parameter_type_module, snake_to_camel(operation_pattern_config.entrypoint) + 'Parameter'
-                        )
-                        operation_pattern_config.parameter_type_cls = parameter_type_cls
+                        if isinstance(operation_pattern_config, OperationHandlerTransactionPatternConfig):
+                            if not operation_pattern_config.entrypoint:
+                                continue
 
-                        _logger.info('Registering storage type')
-                        storage_type_module = importlib.import_module(
-                            f'{self.package}.types.{operation_pattern_config.contract_config.module_name}.storage'
-                        )
-                        storage_type_cls = getattr(
-                            storage_type_module, snake_to_camel(operation_pattern_config.contract_config.module_name) + 'Storage'
-                        )
-                        operation_pattern_config.storage_type_cls = storage_type_cls
+                            _logger.info('Registering parameter type for entrypoint `%s`', operation_pattern_config.entrypoint)
+                            parameter_type_module = importlib.import_module(
+                                f'{self.package}'
+                                f'.types'
+                                f'.{operation_pattern_config.destination_contract_config.module_name}'
+                                f'.parameter'
+                                f'.{camel_to_snake(operation_pattern_config.entrypoint)}'
+                            )
+                            parameter_type_cls = getattr(
+                                parameter_type_module, snake_to_camel(operation_pattern_config.entrypoint) + 'Parameter'
+                            )
+                            operation_pattern_config.parameter_type_cls = parameter_type_cls
+
+                            _logger.info('Registering transaction storage type')
+                            storage_type_module = importlib.import_module(
+                                f'{self.package}.types.{operation_pattern_config.destination_contract_config.module_name}.storage'
+                            )
+                            storage_type_cls = getattr(
+                                storage_type_module,
+                                snake_to_camel(operation_pattern_config.destination_contract_config.module_name) + 'Storage',
+                            )
+                            operation_pattern_config.storage_type_cls = storage_type_cls
+
+                        elif isinstance(operation_pattern_config, OperationHandlerOriginationPatternConfig):
+                            _logger.info('Registering origination storage type')
+                            storage_type_module = importlib.import_module(
+                                f'{self.package}.types.{operation_pattern_config.contract_config.module_name}.storage'
+                            )
+                            storage_type_cls = getattr(
+                                storage_type_module, snake_to_camel(operation_pattern_config.contract_config.module_name) + 'Storage'
+                            )
+                            operation_pattern_config.storage_type_cls = storage_type_cls
 
             elif isinstance(index_config, BigMapIndexConfig):
                 for big_map_handler_config in index_config.handlers:
